title	MS-DOS SERIAL INSTALLABLE DEVICE DRIVER

comment |Installable device driver from Appendix D of "The Personal Computer
from the Inside Out."  The device driver controls a serial printer.  For a
discussion of how the program works, see Sec. 4-6.  This is an example of a
character I/O device driver.
|

;Serial Port Parameters

pnum	equ	0	;Serial port card number
sdata	equ	03f8h	;Serial port transmitter register
mctrl	equ	sdata+4	;Modem control register
status	equ	sdata+5	;Serial port line status register
uparm	equ	0e3h	;Baud rate, word length, and parity
			; Here the port is set for 9600 baud,
			; 8 bits, no parity (see INT 14 section
			; of Technical Reference App. A)
;REQUEST HEADER OFFSETS

len_fld	equ	0	;Length field
ucd_fld	equ	1	;Unit code field
ccd_fld	equ	2	;Command code field
sta_fld	equ	3	;Status field
res_fld	equ	5	;Reserved area field

;Request header additions for initialization

units	equ	13	;Number of units (block devices only)
end_addr	equ	14	;End address for resident code
bpb_ptr	equ	18	;BPB array pointer (block devices only)

;Request header additions for read or write

md	equ	13	;Media descriptor byte (block devices only)
taddr	equ	14	;Data transfer address
count	equ	18	;Number of bytes in buffer
ssn	equ	20	;Starting sector number (block devices only)

cseg	segment public  'code'
	assume	cs:cseg,es:cseg,ds:cseg


;DEVICE HEADER -- must be the first item in an installable device driver

serial	proc  far
next_dev	dd	-1	;Next device header field
attribute	dw	8000h	;Attribute field (char device, no IOCTL)
strategy	dw	dev_strat	;Device strategy routine pointer
interrupt	dw	dev_int	;Device interrupt handler pointer
dev_name	db	'PRN     '	;Eight byte device name field


;DEVICE STRATEGY ROUTINE

dev_strat:	mov	cs:rh_seg,es
	mov	cs:rh_off,bx
	ret

rh_off	dw	?	;Storage for the request header offset
rh_seg	dw	?	; and segment

;JUMP TABLE (for the 13 required device driver functions)

funtab	label	word
	dw	init	;Code 0: initialize device
	dw	media_chk	;Code 1: media changed?
	dw	build_bpb  	;Code 2: build BIOS parm block
	dw	ioctl_in   	;Code 3: input I/O control bytes
	dw	input	;Code 4: input data from I/O device
	dw	nd_input   	;Code 5: nondestructive input, no wait
	dw	in_stat	;Code 6: input device status
	dw	in_flush   	;Code 7: flush (clear) input queue
	dw	output	;Code 8: output data to I/O device
	dw	out_verify 	;Code 9: output data w/verify
	dw	out_stat   	;Code 10: output device status
	dw	out_flush  	;Code 11: flush (clear) output queue
	dw	ioctl_out  	;Code 12: output I/O control bytes


;DEVICE INTERRUPT HANDLER

dev_int:	cld
	push	ds	;Save everything in sight
	push	es
	push	ax
	push	bx
	push	cx
	push	dx
	push	di
	push	si
	mov	ax,cs:rh_seg	;Point ES:BX at request header
	mov	es,ax
	mov	bx,cs:rh_off
	mov	al,es:[bx]+2	;Get cmnd code from request header
	rol	al,1	;Multiply code by 2
	xor	ah,ah
	lea	di,funtab   	;Point to start of jump table
	add	di,ax	;Index into table
	jmp	word ptr [di]	; and jump to requested function

;The following entries are not supported by this device:

media_chk:
build_bpb:
ioctl_in:
input:
nd_input:
in_stat:
in_flush:
ioctl_out:	or	es:word ptr [bx]+sta_fld,8103h
	jmp	exit	;Unknown command, error & quit


;INITIALIZATION (command code 0) - initialize the I/O device

init:	lea	ax,e_o_p	;Store end addr in request header
	mov	es:word ptr [bx]+end_addr,ax
	mov	es:word ptr [bx]+end_addr+2,cs
	mov	ax,uparm	;Get UART parameters
	mov	dx,pnum	;Get serial port card #
	int	14h	;Initialize the UART
	mov	dx,mctrl	;Set DTR and RTS low
	mov	al,3
	out	dx,al
	jmp	exit	;Clean up and return


;OUTPUT (command code 8) - write data bytes to I/O device

output:	mov	si,es:[bx]+taddr	;Get data start address in DS:SI
	mov	ds,es:[bx]+taddr+2
	mov	cx,es:[bx]+count	;Byte count in CX
nxtchr:	mov	ax,[si]	;Get a char
	push	ax
print1:	mov	dx,status	;Read xmitter status
	in	al,dx
	mov	ah,al	;Save in AH
	inc	dx	;Accesses to 8250 must be at least 
	nop		; 2 microseconds apart
	in	al,dx	;Read handshake status
	and	ax,2020h
	xor	ax,2020h
	jnz	print1	;Loop till xmitter and handshake clear
	sub	dx,6	;Then send char in AL
	pop	ax
	out	dx,al
	inc	si	;Point to next data byte
	loop	nxtchr	; and repeat till no data left
	jmp	exit

;OUTPUT WITH VERIFY (command code 9) - write data to I/O device
;and verify by reading data back.  Make same as OUTPUT.

out_verify:	jmp  output

;Just let the following two operations be NOP's for now

out_stat:			;Code 10: output device status
out_flush:			;Code 11: flush (clear) output queue

exit:	or	es:word ptr [bx]+sta_fld,0100H
	pop	si	;Set return status
	pop	di	; and restore all registers
	pop	dx
	pop	cx
	pop	bx
	pop	ax
	pop	es
	pop	ds
	ret

e_o_p:			;End Of Program

serial	endp
cseg	ends
	end	serial
